import argparse
import socket

from webserve.server import Server
from webserve.__version__ import __version__

def find_free_port(start=8010, end=8099):
    ports = list(range(start, end + 1))

    for port in ports:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(('', port))
                return port
            except OSError:
                continue

    return None

def main():
    parser = argparse.ArgumentParser(prog="webserve")
    parser.add_argument("model", type=str, help="Model's HuggingFace path")

    # Server Configuration
    parser.add_argument("--host", type=str, default="0.0.0.0",
                        required=False, help="Server Host Address")
    parser.add_argument("--port", type=int, default=None,
                        required=False, help="Server Port")
    
    # UI Configuration
    parser.add_argument("--darkmode", action="store_true",
                        help="Launch the UI in Dark Mode")
    parser.add_argument("--latex", action="store_true",
                        help="Compile LaTeX in real-time")
    
    # Model Configuration
    parser.add_argument("--context", type=int, default=2048,
                        required=False, help="Context Length (coresponds to vllm's 'max_model_len')")
    parser.add_argument("--temperature", type=float, default=0.6,
                        required=False, help="Model temperature")
    parser.add_argument("--max-tokens", type=int, default=2048,
                        required=False, help="Number of maximum new tokens per generation")

    # Environment Configuration
    parser.add_argument("--gpu-count", type=int, default=1,
                        required=False, help="Number of GPUs to use")

    # Other
    parser.add_argument("--version", action="version",
                        version=f"\033[34m(VERSION)\033[0m WebServe version {__version__}")

    args = parser.parse_args()

    print(f"\033[34m(NOTICE)\033[0m Launching WebServe version {__version__}...")

    port = args.port

    if port is None:
        print("\033[34m(Preparing server...)\033[0m Looking for port to bind to. Avoid this by specifying port with --port PORT.")
        port = find_free_port()
        if port is None:
            print("\033[31m(ERROR)\033[0m Failed to find port to run the server. Please use --port PORT.")
            return
        else:
            print(f"\033[34m(NOTICE)\033[0m Selected port {port}.")

    print("\033[34m(Preparing server...)\033[0m Starting vLLM engine.")

    server = Server(
        args.model,
        gpu_count=args.gpu_count,
        max_model_len=args.context,
        max_tokens=args.max_tokens,
        temperature=args.temperature,
        latex=args.latex,
        darkmode=args.darkmode,
        verbose=False
    )

    print(f"\033[34m(NOTICE)\033[0m vLLM Engine ready.")
    print("\033[34m(Preparing server...)\033[0m Starting HTTP server.")
    
    server.listen(port, host=args.host, daemon=True)

    print(f"\033[34m(NOTICE)\033[0m Server ready.\n")
    print(f"\033[34m(CONFIG)\033[0m Model: {args.model}")
    print(f"\033[34m(CONFIG)\033[0m Temperature: {args.temperature}")
    print(f"\033[34m(CONFIG)\033[0m Context Length: {args.context} tok.")
    print(f"\033[34m(CONFIG)\033[0m Max New Tokens: {args.max_tokens} tok.")
    print(f"\033[34m(CONFIG)\033[0m GPU count: {args.gpu_count}")
    print(f"\033[34m(CONFIG)\033[0m LaTeX: {'ON' if args.latex else 'OFF'}")
    print(f"\033[34m(CONFIG)\033[0m WebServe listening on HTTP port {port}.")
    print(f"\033[34m(CONFIG)\033[0m http://localhost:{port}/\n")
    print("Type 'exit' to close server.")
    
    while input() != 'exit':
        continue

if __name__ == "__main__":
    main()


import json
import os
import threading

import fastapi
from transformers import AutoTokenizer
import uvicorn

from webserve.__version__ import __version__

class Server:
    _STATIC_FILES = ["index.html", "main.css", "display.js", "stream.js"]

    def __init__(
        self,
        model: str,
        *,
        gpu_count: int | None = 1,
        max_model_len: int | None = 2048,
        max_tokens: int | None = 2048,
        temperature: float | None = 0.6,
        latex: bool | None = False,
        darkmode: bool | None = False,
        verbose: bool | None = False,
    ):
        from vllm import SamplingParams
        from vllm.engine.arg_utils import AsyncEngineArgs
        from vllm.v1.engine.async_llm import AsyncLLM

        self.darkmode = darkmode
        self.verbose = verbose
        self.latex = latex

        self._tokenizer = AutoTokenizer.from_pretrained(model)

        self._model = model
        self._temperature = temperature
        self._id_status = 0

        engine_args = AsyncEngineArgs(
            model=model,
            gpu_memory_utilization=0.9,
            max_model_len=max_model_len,
            tensor_parallel_size=gpu_count,
            enforce_eager=True, # load faster
        )

        self._engine = AsyncLLM.from_engine_args(engine_args)

        self._samp = SamplingParams(
            temperature=temperature,
            max_tokens=max_tokens,
        )

        self._app = fastapi.FastAPI()
        self._static = self._load_static()
        self._setup_routes()

    def _get_id(self) -> str:
        id_ = f"vllm-request-{self._id_status}"
        self._id_status += 1
        return id_

    def _get_actual_query(self, messages: list) -> str:
        return self._tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=True,
        )

    def _setup_routes(self) -> None:
        @self._app.get("/", response_class=fastapi.responses.HTMLResponse)
        async def get_index():
            return self._static["index.html"]
        
        @self._app.get("/version")
        async def get_main_css():
            content = f"webserve v{__version__}"
            return fastapi.responses.Response(content, media_type="text/plain")

        @self._app.get("/robots.txt")
        async def get_robots_txt():
            content = "User-agent: *\nDisallow: /"
            return fastapi.responses.Response(content, media_type="text/plain")

        @self._app.get("/main.css")
        async def get_main_css():
            return fastapi.responses.Response(self._static["main.css"], media_type="text/css")

        @self._app.get("/display.js")
        async def get_display_js():
            return fastapi.responses.Response(self._static["display.js"], media_type="application/javascript")
        
        @self._app.get("/stream.js")
        async def get_stream_js():
            return fastapi.responses.Response(self._static["stream.js"], media_type="application/javascript")
        
        @self._app.get("/config.js")
        async def get_config_js():
            return fastapi.responses.Response(self._get_config(), media_type="application/javascript")

        @self._app.websocket("/v1/retrieveResponse")
        async def websocket_endpoint(websocket: fastapi.WebSocket):
            await websocket.accept()
            try:
                incomming = await websocket.receive_text()
                data = json.loads(incomming)
                query = self._get_actual_query(data)
                if query:
                    text = None
                    async for output in self._engine.generate(request_id=self._get_id(), prompt=query, sampling_params=self._samp):
                        text = output.outputs[0].text

                        await websocket.send_json({
                            "completed": False,
                            "output": text,
                            "error": None
                        })

                        if output.finished: 
                            await websocket.send_json({
                                "completed": True,
                                "output": text,
                                "error": None
                            })
                            break

            except fastapi.WebSocketDisconnect:
                pass
            except Exception as e:
                await websocket.send_json({
                    "completed": False,
                    "output": "",
                    "error": str(e)
                })
                print(f"\033[31m(ERROR)\033[0m {e}")
            finally:
                await websocket.close()

    def _load_static(self) -> dict[str, str]:
        current_directory = os.path.dirname(os.path.abspath(__file__))
        static_path = os.path.join(current_directory, "static/")
        output = {}
        for filename in self._STATIC_FILES:
            with open(os.path.join(static_path, filename), 'r') as file:
                output[filename] = file.read()
        return output
    
    def _get_config(self) -> str:
        config = {
            "mode": 'dark' if self.darkmode else 'light',
            "model": self._model,
            "temperature": self._temperature,
            "latex": self.latex
        }
        return f"const config = {json.dumps(config, indent=4)}"

    def _run(self, host: str, port: int) -> None:
        config = uvicorn.Config(
            self._app, 
            host=host, 
            port=port, 
            log_level=('info' if self.verbose else 'critical')
        )
        server = uvicorn.Server(config)
        server.run()

    def listen(
        self, 
        port: int = 80, 
        *, 
        host: str | None = "0.0.0.0",
        daemon: bool | None = False
    ) -> None:
        if daemon:
            server_thread = threading.Thread(target=self._run, args=(host, port), daemon=True)
            server_thread.start()
        else:
            self._run(host, port)